"""
Helper to wrap watchdog.
"""
import warnings
import os.path
import sys
import logging
import threading
import weakref
from typing import List, Tuple, Optional, Set, Sequence, Any
from robocorp_ls_core.uris import normalize_drive

log = logging.getLogger(__name__)

__file__ = os.path.abspath(__file__)  # @ReservedAssignment

# Hack so that we don't break the runtime on versions prior to Python 3.8.
if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass


else:
    from typing import Protocol


class PathInfo(object):
    __slots__ = ["path", "recursive"]

    def __init__(self, path: str, recursive: bool):
        path = str(path)
        self.path = normalize_drive(path)
        self.recursive = recursive

    def __eq__(self, o):
        if isinstance(o, PathInfo):
            return o.path == self.path and o.recursive == self.recursive

        return False

    def __ne__(self, o):
        return not self == o

    def __hash__(self, *args, **kwargs):
        return hash(tuple(self.path, self.recursive))

    def __str__(self):
        return f"PathInfo[{self.path}, {self.recursive}]"

    __repr__ = __str__


class IFSCallback(Protocol):
    def __call__(self, src_path, *call_args):
        pass


class IFSWatch(Protocol):
    def stop_tracking(self):
        pass


class IFSObserver(Protocol):
    def notify_on_any_change(
        self,
        paths: List[PathInfo],
        on_change: IFSCallback,
        call_args=(),
        extensions: Optional[Sequence[str]] = None,
    ) -> IFSWatch:
        pass

    def dispose(self):
        pass


def _get_watchdog_lib_dir():
    parent_dir = os.path.dirname(__file__)
    watchdog_dir = os.path.join(parent_dir, "libs", "watchdog_lib")
    if not os.path.exists(watchdog_dir):
        raise RuntimeError("Expected: %s to exist." % (watchdog_dir,))
    return watchdog_dir


def _import_watchdog():
    warnings.filterwarnings("ignore", message=".*failed to import fsevents.*")

    try:
        import watchdog
    except ImportError:
        sys.path.append(_get_watchdog_lib_dir())
        import watchdog  # @UnusedImport


def _get_fsnotify_lib_dir():
    parent_dir = os.path.dirname(__file__)
    fsnotify_dir = os.path.join(parent_dir, "libs", "fsnotify_lib")
    if not os.path.exists(fsnotify_dir):
        msg = "Expected: %s to exist.\nDetails:\n" % (fsnotify_dir,)
        check = fsnotify_dir
        while True:
            dirname = os.path.dirname(check)
            exists = os.path.exists(dirname)
            msg += f"{dirname} exists: {exists}\n"
            if exists:
                msg += f"{dirname} contents: {os.listdir(dirname)}\n"

            if not dirname or dirname == check or exists:
                break
            check = dirname

        raise RuntimeError(msg)
    return fsnotify_dir


def _import_fsnotify():
    try:
        import fsnotify
    except ImportError:
        sys.path.append(_get_fsnotify_lib_dir())
        import fsnotify  # @UnusedImport


class _Notifier(threading.Thread):
    def __init__(
        self, callback, timeout: int, extensions: Optional[Tuple[str, ...]] = None
    ):
        """
        :param callback:
            Callable which should be called when a file with the given extension changes.
        :param timeout:
            The amount of time which should elapse to send notifications (so, multiple
            modifications of the same file will be sent as a single notification).
        :param extensions:
            Only notify file changes of the given extensions.
        """
        threading.Thread.__init__(self)
        self.name = "FS Notifier Thread (_Notifier class)"
        self.daemon = True

        self._changes: Set[tuple] = set()
        self._event = threading.Event()
        self._timeout = timeout
        self._disposed = False
        self._callback = callback
        self._extensions = extensions

    def run(self):
        import time

        while not self._disposed:
            self._event.wait()
            time.sleep(self._timeout)
            if self._disposed:
                return

            changes = self._changes
            self._changes = set()
            for src_path_and_call_args in changes:
                try:
                    self._callback(*src_path_and_call_args)
                except:
                    log.exception(
                        "Error handling change on: %s", src_path_and_call_args
                    )
            self._event.clear()

    def on_change(self, src_path, *call_args):
        if self._extensions:
            for ext in self._extensions:
                if src_path.lower().endswith(ext):
                    # Ok, proceed to notify.
                    break
            else:
                # Skip this notification as it's not in one of the available
                # extensions.
                return
        src_path = normalize_drive(src_path)
        self._changes.add((src_path,) + tuple(call_args))
        self._event.set()

    def dispose(self):
        self._disposed = True
        self._event.set()


def create_notifier(callback, timeout, extensions=None):
    notifier = _Notifier(callback, timeout, extensions)
    notifier.start()
    return notifier


class _DummyWatchList(object):
    def stop_tracking(self):
        pass

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IFSWatch = check_implements(self)


class _FSNotifyWatchList(object):
    def __init__(self, new_tracked_paths, new_notifications, observer):
        self._new_tracked_paths = new_tracked_paths
        self._new_notifications = new_notifications
        self._observer = weakref.ref(observer)

    def stop_tracking(self):
        observer = self._observer()
        if observer is not None and self._new_tracked_paths:
            observer._stop_tracking(self._new_tracked_paths, self._new_notifications)
        self._new_tracked_paths = []
        self._new_notifications = []

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IFSWatch = check_implements(self)


class _FSNotifyObserver(threading.Thread):
    def __init__(self, extensions: Optional[Tuple[str, ...]]):
        from robocorp_ls_core import load_ignored_dirs

        threading.Thread.__init__(self)
        import fsnotify

        self.name = "_FSNotifyObserver"
        self.daemon = True

        self._disposed = threading.Event()

        if extensions is None:
            extensions = ()
        else:
            extensions = tuple(extensions)

        watcher = self._watcher = fsnotify.Watcher()
        poll_time_str: Optional[str] = os.environ.get("ROBOTFRAMEWORK_LS_POLL_TIME")
        watcher.target_time_for_notification = 4.0
        watcher.target_time_for_single_scan = 4.0

        if poll_time_str:
            try:
                poll_time = int(poll_time_str)
            except Exception:
                log.exception(
                    "Unable to convert ROBOTFRAMEWORK_LS_POLL_TIME (%s) to an int.",
                    poll_time,
                )
            else:
                watcher.target_time_for_notification = poll_time
                watcher.target_time_for_single_scan = poll_time

        watcher.accepted_file_extensions = extensions
        watcher.accept_directory = load_ignored_dirs.create_accept_directory_callable()

        self._all_paths_to_track: List[fsnotify.TrackedPath] = []
        self._lock = threading.Lock()
        self._notifications: List[Tuple[str, Any, List[Any]]] = []
        self._was_started = False

    def dispose(self):
        if not self._disposed.is_set():
            self._disposed.set()
            self._watcher.dispose()

    def run(self):
        log.debug("Started listening on _FSNotifyObserver.")
        try:
            while not self._disposed.is_set():
                for _change, src_path in self._watcher.iter_changes():
                    lower = src_path.lower()
                    for path, on_change, call_args in self._notifications:
                        if lower.startswith(path.lower()):
                            on_change(normalize_drive(src_path), *call_args)
        except:
            log.exception("Error collecting changes in _FSNotifyObserver.")
        finally:
            log.debug("Finished listening on _FSNotifyObserver.")

    def _tracked_paths_set_on_thread(self):
        with self._lock:
            all_paths_to_track = self._all_paths_to_track[:]
        self._watcher.set_tracked_paths(all_paths_to_track)

    def _stop_tracking(self, new_paths_to_track, new_notifications):
        with self._lock:
            for path in new_paths_to_track:
                self._all_paths_to_track.remove(path)
            for notification in new_notifications:
                self._notifications.remove(notification)
        threading.Thread(target=self._tracked_paths_set_on_thread).start()

    def notify_on_any_change(
        self,
        paths: List[PathInfo],
        on_change: IFSCallback,
        call_args=(),
        extensions: Optional[Sequence[str]] = None,
    ) -> IFSWatch:
        if self._disposed.is_set():
            return _DummyWatchList()

        import fsnotify

        used_on_change = on_change
        if extensions:

            def on_change_with_extensions(src_path, *args):
                lower = src_path.lower()
                if lower.endswith(extensions):
                    on_change(src_path, *args)

            used_on_change = on_change_with_extensions

        with self._lock:
            new_paths_to_track = []
            new_notifications = []
            for path in paths:
                tracked_path = fsnotify.TrackedPath(path.path, path.recursive)
                new_paths_to_track.append(tracked_path)
                new_notifications.append((path.path, used_on_change, call_args))

            self._notifications.extend(new_notifications)
            self._all_paths_to_track.extend(new_paths_to_track)

        threading.Thread(target=self._tracked_paths_set_on_thread).start()
        if not self._was_started:
            with self._lock:
                # Check again with the lock in place.
                if not self._was_started:
                    self._was_started = True
                    self.start()

        return _FSNotifyWatchList(new_paths_to_track, new_notifications, self)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IFSObserver = check_implements(self)


class _WatchdogWatchList(object):
    def __init__(self, watches, observer):
        self.watches = watches
        self._observer = weakref.ref(observer)

    def stop_tracking(self):
        observer = self._observer()
        if observer is not None:
            for watch in self.watches:
                try:
                    observer.unschedule(watch)
                except Exception:
                    pass
        self.watches = []

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IFSWatch = check_implements(self)


class _WatchdogObserver(object):
    def __init__(self, extensions=None):
        from watchdog.observers import Observer

        self._observer = Observer()
        self._started = False
        self._extensions = extensions

    def dispose(self):
        self._observer.stop()

    def _start(self):
        if not self._started:
            self._started = True
            self._observer.start()

    def notify_on_any_change(
        self,
        paths: List[PathInfo],
        on_change: IFSCallback,
        call_args=(),
        extensions: Optional[Sequence[str]] = None,
    ) -> IFSWatch:
        """
        To be used as:

        notifier = create_notifier(callback=on_file_change, timeout=0.5)

        observer = create_observer()

        watch = observer.notify_on_any_change(
            [PathInfo('a', recursive=True)],
            notifier.on_change,
            extensions=('.py', '.robot'),
        )
        ...
        watch.stop_tracking()

        notifier.dispose()
        observer.dispose()

        Multiple changes on the same file will be sent as a single change (if
        the changes occur during the available timeout).

        :param list(PathInfo) paths:
        :param list(str) extensions:
            The file extensions that should be tracked.
        """
        from watchdog.events import FileSystemEventHandler

        if not extensions:
            extensions = self._extensions

        class _Handler(FileSystemEventHandler):
            def __init__(
                self,
            ):
                FileSystemEventHandler.__init__(self)

            def on_any_event(self, event):
                if extensions is not None:
                    for ext in extensions:
                        if event.src_path.endswith(ext):
                            break
                    else:
                        return
                # Note: notify on directory and file changes.
                on_change(event.src_path, *call_args)
                try:
                    dest_path = event.dest_path
                except AttributeError:
                    pass
                else:
                    if dest_path:
                        on_change(dest_path, *call_args)

        handler = _Handler()
        watches = []
        for path_info in paths:
            watches.append(
                self._observer.schedule(
                    handler, path_info.path, recursive=path_info.recursive
                )
            )

        self._start()
        return _WatchdogWatchList(watches, self._observer)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IFSObserver = check_implements(self)


class _DummyFSObserver(object):
    def notify_on_any_change(
        self,
        paths: List[PathInfo],
        on_change: IFSCallback,
        call_args=(),
        extensions: Optional[Sequence[str]] = None,
    ) -> IFSWatch:
        return _DummyWatchList()

    def dispose(self):
        pass

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IFSObserver = check_implements(self)


def create_observer(backend: str, extensions: Optional[Tuple[str, ...]]) -> IFSObserver:
    """
    :param backend:
        The backend to use.
        'fsnotify', 'watchdog' or 'dummy'.
    """
    if backend == "watchdog":
        _import_watchdog()
        return _WatchdogObserver(extensions)

    elif backend == "fsnotify":
        _import_fsnotify()
        return _FSNotifyObserver(extensions)

    elif backend == "dummy":
        return _DummyFSObserver()

    raise AssertionError(f"Unhandled observer: {backend}")


def create_remote_observer(
    backend: str, extensions: Optional[Tuple[str, ...]]
) -> IFSObserver:
    """
    :param backend:
        The backend to use.
        'fsnotify' or 'watchdog'.
    """
    assert backend in ("watchdog", "fsnotify")
    from robocorp_ls_core.remote_fs_observer_impl import RemoteFSObserver

    return RemoteFSObserver(backend, extensions)



from typing import List

from robocorp_ls_core.lsp import FoldingRangeTypedDict
from robotframework_ls.impl.protocols import ICompletionContext
from robocorp_ls_core.robotframework_log import get_logger

log = get_logger(__name__)


def folding_range(
    completion_context: ICompletionContext,
) -> List[FoldingRangeTypedDict]:
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.protocols import NodeInfo

    ast = completion_context.get_ast()
    completion_context.check_cancelled()

    ret: List[FoldingRangeTypedDict] = []
    node: NodeInfo

    # i.e.: any node that spans more than one line
    # should be added to the result.
    for node in ast_utils.iter_all_nodes(ast):
        completion_context.check_cancelled()
        try:
            start_line = node.node.lineno - 1
            end_line = node.node.end_lineno - 1
            if end_line > start_line:
                ret.append({"startLine": start_line, "endLine": end_line})
        except Exception:
            log.exception("Error computing range")

    return ret



def get_return():
    return 10


def generator():
    print('start')  # break here
    yield 10  # step 1

    return \
            get_return()  # step 2


if __name__ == '__main__':
    for i in generator():  # generator return
        print(i)

    print('TEST SUCEEDED!')  # step 3





def Call():
    variable_for_test_1 = 10  # Break here
    variable_for_test_2 = 20
    variable_for_test_3 = {'a':30, 'b':20}
    locals()[u'\u16A0'] = u'\u16A1'  # unicode variable (would be syntax error on py2).

    all_vars_set = True  # Break 2 here


if __name__ == '__main__':
    Call()
    print('TEST SUCEEDED!')



# encoding: utf-8
"""
This script tests ``PathSpec``.
"""

import unittest

import pathspec


class PathSpecTest(unittest.TestCase):
	"""
	The ``PathSpecTest`` class tests the ``PathSpec`` class.
	"""

	def test_01_absolute_dir_paths_1(self):
		"""
		Tests that absolute paths will be properly normalized and matched.
		"""
		spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'foo',
		])
		results = set(spec.match_files([
			'/a.py',
			'/foo/a.py',
			'/x/a.py',
			'/x/foo/a.py',
			'a.py',
			'foo/a.py',
			'x/a.py',
			'x/foo/a.py',
		]))
		self.assertEqual(results, {
			'/foo/a.py',
			'/x/foo/a.py',
			'foo/a.py',
			'x/foo/a.py',
		})

	def test_01_absolute_dir_paths_2(self):
		"""
		Tests that absolute paths will be properly normalized and matched.
		"""
		spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'/foo',
		])
		results = set(spec.match_files([
			'/a.py',
			'/foo/a.py',
			'/x/a.py',
			'/x/foo/a.py',
			'a.py',
			'foo/a.py',
			'x/a.py',
			'x/foo/a.py',
		]))
		self.assertEqual(results, {
			'/foo/a.py',
			'foo/a.py',
		})

	def test_01_current_dir_paths(self):
		"""
		Tests that paths referencing the current directory will be properly
		normalized and matched.
		"""
		spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'*.txt',
			'!test1/',
		])
		results = set(spec.match_files([
			'./src/test1/a.txt',
			'./src/test1/b.txt',
			'./src/test1/c/c.txt',
			'./src/test2/a.txt',
			'./src/test2/b.txt',
			'./src/test2/c/c.txt',
		]))
		self.assertEqual(results, {
			'./src/test2/a.txt',
			'./src/test2/b.txt',
			'./src/test2/c/c.txt',
		})

	def test_01_match_files(self):
		"""
		Tests that matching files one at a time yields the same results as
		matching multiples files at once.
		"""
		spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'*.txt',
			'!test1/',
		])
		test_files = [
			'src/test1/a.txt',
			'src/test1/b.txt',
			'src/test1/c/c.txt',
			'src/test2/a.txt',
			'src/test2/b.txt',
			'src/test2/c/c.txt',
		]
		single_results = set(filter(spec.match_file, test_files))
		multi_results = set(spec.match_files(test_files))
		self.assertEqual(single_results, multi_results)

	def test_01_windows_current_dir_paths(self):
		"""
		Tests that paths referencing the current directory will be properly
		normalized and matched.
		"""
		spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'*.txt',
			'!test1/',
		])
		results = set(spec.match_files([
			'.\\src\\test1\\a.txt',
			'.\\src\\test1\\b.txt',
			'.\\src\\test1\\c\\c.txt',
			'.\\src\\test2\\a.txt',
			'.\\src\\test2\\b.txt',
			'.\\src\\test2\\c\\c.txt',
		], separators=('\\',)))
		self.assertEqual(results, {
			'.\\src\\test2\\a.txt',
			'.\\src\\test2\\b.txt',
			'.\\src\\test2\\c\\c.txt',
		})

	def test_01_windows_paths(self):
		"""
		Tests that Windows paths will be properly normalized and matched.
		"""
		spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'*.txt',
			'!test1/',
		])
		results = set(spec.match_files([
			'src\\test1\\a.txt',
			'src\\test1\\b.txt',
			'src\\test1\\c\\c.txt',
			'src\\test2\\a.txt',
			'src\\test2\\b.txt',
			'src\\test2\\c\\c.txt',
		], separators=('\\',)))
		self.assertEqual(results, {
			'src\\test2\\a.txt',
			'src\\test2\\b.txt',
			'src\\test2\\c\\c.txt',
		})

	def test_02_eq(self):
		"""
		Tests equality.
		"""
		first_spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'*.txt',
			'!test1/',
		])
		second_spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'*.txt',
			'!test1/',
		])
		self.assertEqual(first_spec, second_spec)

	def test_02_ne(self):
		"""
		Tests equality.
		"""
		first_spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'*.txt',
		])
		second_spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'!*.txt',
		])
		self.assertNotEqual(first_spec, second_spec)

	def test_01_addition(self):
		"""
		Test pattern addition using + operator
		"""
		first_spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'test.txt',
			'test.png'
		])
		second_spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'test.html',
			'test.jpg'
		])
		combined_spec = first_spec + second_spec
		results = set(combined_spec.match_files([
			'test.txt',
			'test.png',
			'test.html',
			'test.jpg'
		], separators=('\\',)))
		self.assertEqual(results, {
			'test.txt',
			'test.png',
			'test.html',
			'test.jpg'
		})

	def test_02_addition(self):
		"""
		Test pattern addition using += operator
		"""
		spec = pathspec.PathSpec.from_lines('gitwildmatch', [
			'test.txt',
			'test.png'
		])
		spec += pathspec.PathSpec.from_lines('gitwildmatch', [
			'test.html',
			'test.jpg'
		])
		results = set(spec.match_files([
			'test.txt',
			'test.png',
			'test.html',
			'test.jpg'
		], separators=('\\',)))
		self.assertEqual(results, {
			'test.txt',
			'test.png',
			'test.html',
			'test.jpg'
		})



from typing import List

from robocorp_ls_core.lsp import CodeLensTypedDict, RangeTypedDict, PositionTypedDict
from robotframework_ls.impl.protocols import ICompletionContext, NodeInfo
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.protocols import ITestInfoTypedDict
from robocorp_ls_core.lsp import CommandTypedDict
from robotframework_ls.impl.robot_constants import ROBOT_FILE_EXTENSIONS

log = get_logger(__name__)


def list_tests(completion_context: ICompletionContext) -> List[ITestInfoTypedDict]:
    from robot.api import Token
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.ast_utils import create_range_from_token

    ast = completion_context.get_ast()
    completion_context.check_cancelled()

    ret: List[ITestInfoTypedDict] = []
    node: NodeInfo
    for node in ast_utils.iter_tests(ast):
        completion_context.check_cancelled()
        try:

            test_case_name_token = node.node.header.get_token(Token.TESTCASE_NAME)
            if not test_case_name_token:
                continue

            ret.append(
                {
                    "uri": completion_context.doc.uri,
                    "path": completion_context.doc.path,
                    "name": test_case_name_token.value,
                    "range": create_range_from_token(test_case_name_token),
                }
            )
        except Exception:
            log.exception("Error listing tests in document.")

    return ret


def code_lens_runs(completion_context: ICompletionContext) -> List[CodeLensTypedDict]:
    from robot.api import Token
    from robotframework_ls.impl.ast_utils import create_range_from_token

    ast = completion_context.get_ast()
    completion_context.check_cancelled()

    ret: List[CodeLensTypedDict] = []

    start: PositionTypedDict = {"line": 0, "character": 0}
    end: PositionTypedDict = {"line": 0, "character": 0}
    code_lens_range: RangeTypedDict = {"start": start, "end": end}

    sections = ast.sections

    test_case_sections = [
        x for x in sections if x.__class__.__name__ == "TestCaseSection"
    ]

    if len(test_case_sections) > 0:
        # Run Test command
        command: CommandTypedDict = {
            "title": "Run Suite",
            "command": "robot.runTest",
            "arguments": [
                {
                    "uri": completion_context.doc.uri,
                    "path": completion_context.doc.path,
                    "name": "*",
                }
            ],
        }

        ret.append({"range": code_lens_range, "command": command})

        # Debug Test command
        command = {
            "title": "Debug Suite",
            "command": "robot.debugTest",
            "arguments": [
                {
                    "uri": completion_context.doc.uri,
                    "path": completion_context.doc.path,
                    "name": "*",
                }
            ],
        }

        ret.append({"range": code_lens_range, "command": command})

    for test_case in test_case_sections:
        try:
            for test_node in test_case.body:
                header = getattr(test_node, "header", None)
                if not header:
                    continue
                test_case_name_token = header.get_token(Token.TESTCASE_NAME)
                if not test_case_name_token:
                    continue

                completion_context.check_cancelled()

                code_lens_range = create_range_from_token(test_case_name_token)

                # Run Test command
                command = {
                    "title": "Run",
                    "command": "robot.runTest",
                    "arguments": [
                        {
                            "uri": completion_context.doc.uri,
                            "path": completion_context.doc.path,
                            "name": test_case_name_token.value,
                        }
                    ],
                }

                code_lens_dct: CodeLensTypedDict = {
                    "range": code_lens_range,
                    "command": command,
                }
                ret.append(code_lens_dct)

                # Debug Test command
                command = {
                    "title": "Debug",
                    "command": "robot.debugTest",
                    "arguments": [
                        {
                            "uri": completion_context.doc.uri,
                            "path": completion_context.doc.path,
                            "name": test_node.name,
                        }
                    ],
                }

                code_lens_dct = {"range": code_lens_range, "command": command}
                ret.append(code_lens_dct)
        except Exception:
            log.exception("Error computing code lens")

    return ret


TEST_CASE_HEADER = "*** Test Case ***\n"


def _iter_rf_interactive_items(ast):
    from robot.api import Token

    sections = ast.sections

    for section in sections:
        if section.__class__.__name__ == "TestCaseSection":
            for node in section.body:
                header = getattr(node, "header", None)
                if header:
                    name_token = header.get_token(Token.TESTCASE_NAME)
                    if name_token:
                        yield name_token, TEST_CASE_HEADER, node

        elif section.__class__.__name__ == "KeywordSection":
            for node in section.body:
                header = getattr(node, "header", None)
                if header:
                    name_token = header.get_token(Token.KEYWORD_NAME)
                    if name_token:
                        yield name_token, "*** Keyword ***\n", node

        elif section.__class__.__name__ == "SettingSection":
            header = getattr(section, "header", None)
            if header:
                name_token = header.get_token(Token.SETTING_HEADER)
                # The header is not needed since it's already a part of its tokens
                # (unlike test cases/keywords)
                if name_token:
                    yield name_token, "", section

        elif section.__class__.__name__ == "VariableSection":
            header = getattr(section, "header", None)
            if header:
                name_token = header.get_token(Token.VARIABLE_HEADER)
                # The header is not needed since it's already a part of its tokens
                # (unlike test cases/keywords)
                if name_token:
                    yield name_token, "", section


def code_lens_rf_interactive(
    completion_context: ICompletionContext,
) -> List[CodeLensTypedDict]:
    from robotframework_ls import import_rf_interactive

    import_rf_interactive()

    ast = completion_context.get_ast()
    completion_context.check_cancelled()

    ret: List[CodeLensTypedDict] = []

    for name_token, header, node in _iter_rf_interactive_items(ast):
        completion_context.check_cancelled()
        ret.append(
            _create_rf_interactive_code_lens(
                completion_context, name_token, header, node
            )
        )

    return ret


def _create_rf_interactive_code_lens(
    completion_context, token, header, node
) -> CodeLensTypedDict:
    from robotframework_ls.impl.ast_utils import create_range_from_token

    code_lens_range = create_range_from_token(token)

    code_lens_dct: CodeLensTypedDict = {
        "range": code_lens_range,
        # Note: initially the command was None and would be resolved later on,
        # but the experience on VSCode appeared sluggish, so, it was changed to
        # always return the command from the start.
        #
        # i.e.: VSCode took too much time to actually do the resolve in some
        # background thread and in the meanwhile it wouldn't show (which makes
        # for a bad experience -- I thought it'd always be there and would
        # resolve just when clicked, but that's not the case).
        # This makes our code-lenses messages bigger, but the processing itself
        # isn't such a big problem and the final experience is better.
        #
        # The code that did the resolution is still there, so, if VSCode
        # improves we can just not return the command here so that it's
        # lazily computed.
        "command": _code_lens_rf_interactive_command(
            header, node, completion_context.doc.uri
        ),
        "data": {"uri": completion_context.doc.uri, "type": "rf_interactive"},
    }
    return code_lens_dct


def _code_lens_rf_interactive_command(header, node, uri) -> CommandTypedDict:
    from robotframework_interactive.ast_to_code import ast_to_code

    code_lens_command: CommandTypedDict = {
        "title": "Run in Interactive Console"
        if header == TEST_CASE_HEADER
        else "Load in Interactive Console",
        "command": "robot.interactiveShell",
        "arguments": [{"code": header + ast_to_code(node), "uri": uri}],
    }
    return code_lens_command


def code_lens_resolve(
    completion_context: ICompletionContext, code_lens: CodeLensTypedDict
):
    # Fill in the command arguments
    command = code_lens.get("command")
    data = code_lens.get("data")
    if (
        command is None
        and isinstance(data, dict)
        and data.get("type") == "rf_interactive"
    ):
        from robotframework_ls import import_rf_interactive

        import_rf_interactive()

        code_lens_range: RangeTypedDict = code_lens["range"]
        code_lens_line = code_lens_range["start"]["line"]

        ast = completion_context.get_ast()
        for name_token, header, node in _iter_rf_interactive_items(ast):
            if name_token.lineno - 1 == code_lens_line:

                code_lens["command"] = _code_lens_rf_interactive_command(
                    header, node, data["uri"]
                )
                break
        else:
            # Unable to resolve
            log.info("Unable to resolve code lens.")
            return code_lens

    return code_lens


def code_lens(completion_context: ICompletionContext) -> List[CodeLensTypedDict]:
    import os
    from robotframework_ls.impl.robot_lsp_constants import OPTION_ROBOT_SHOW_CODE_LENSES

    config = completion_context.config
    if config is not None and not config.get_setting(
        OPTION_ROBOT_SHOW_CODE_LENSES, bool, True
    ):
        return []

    path = completion_context.doc.path
    if not path or not os.path.exists(os.path.dirname(path)):
        # The cwd must exist.
        return []
    if not path.endswith(ROBOT_FILE_EXTENSIONS):
        return []

    code_lenses = code_lens_runs(completion_context)
    code_lenses.extend(code_lens_rf_interactive(completion_context))
    return code_lenses



from typing import Optional

from robocorp_ls_core.cache import instance_cache
from robocorp_ls_core.protocols import check_implements
from robotframework_ls.impl.protocols import (
    VariableKind,
    IVariableFound,
    ICompletionContext,
)


class VariableFoundFromToken(object):
    def __init__(
        self,
        completion_context,
        variable_token,
        variable_value,
        variable_name=None,
        variable_kind=VariableKind.VARIABLE,
    ):
        self.completion_context = completion_context
        self.variable_token = variable_token

        if variable_name is None:
            variable_name = variable_token.value

        self.variable_name = variable_name
        if isinstance(variable_value, (list, tuple, set)):
            if len(variable_value) == 1:
                self.variable_value = str(next(iter(variable_value)))
            else:
                self.variable_value = str(variable_value)
        else:
            self.variable_value = str(variable_value)
        self.variable_kind = variable_kind

    @property  # type: ignore
    @instance_cache
    def source(self):
        from robocorp_ls_core import uris

        return uris.to_fs_path(self.completion_context.doc.uri)

    @property
    def lineno(self):
        return self.variable_token.lineno - 1  # Make 0-based

    @property
    def end_lineno(self):
        return self.variable_token.lineno - 1  # Make 0-based

    @property
    def col_offset(self):
        return self.variable_token.col_offset

    @property
    def end_col_offset(self):
        return self.variable_token.end_col_offset

    def __typecheckself__(self) -> None:
        _: IVariableFound = check_implements(self)

    def __str__(self):
        return f"{self.__class__.__name__}({self.variable_name})"

    __repr__ = __str__


class VariableFoundFromPythonAst(object):
    def __init__(
        self,
        path: str,
        lineno: int,
        col: int,
        end_lineno: int,
        end_col: int,
        variable_value: str,
        variable_name: str,
    ):
        self.lineno = lineno
        self.col_offset = col
        self.end_lineno = end_lineno
        self.end_col_offset = end_col

        self.completion_context: Optional[ICompletionContext] = None
        self._path = path
        self.variable_name = variable_name
        self.variable_value = variable_value
        self.variable_kind = VariableKind.PYTHON

    @property
    def source(self):
        return self._path

    def __typecheckself__(self) -> None:
        _: IVariableFound = check_implements(self)

    def __str__(self):
        return f"{self.__class__.__name__}({self.variable_name})"

    __repr__ = __str__


class VariableFoundFromSettings(object):
    variable_kind = VariableKind.SETTINGS

    def __init__(self, variable_name, variable_value, source="", lineno=0):
        self.completion_context = None
        self.variable_name = variable_name
        self.variable_value = str(variable_value)
        self._source = source
        self._lineno = lineno

    @property
    def source(self):
        return self._source

    @property
    def lineno(self):
        return self._lineno

    @property
    def end_lineno(self):
        return self._lineno

    @property
    def col_offset(self):
        return 0

    @property
    def end_col_offset(self):
        return 0

    def __typecheckself__(self) -> None:
        _: IVariableFound = check_implements(self)

    def __str__(self):
        return f"{self.__class__.__name__}({self.variable_name})"

    __repr__ = __str__


class VariableFoundFromBuiltins(VariableFoundFromSettings):
    variable_kind = VariableKind.BUILTIN


class VariableFoundFromYaml(VariableFoundFromSettings):
    variable_kind = VariableKind.YAML


class VariableFoundFromArgumentsFile(VariableFoundFromSettings):
    variable_kind = VariableKind.ARGUMENTS_FILE



from robocorp_ls_core.basic import isinstance_name
from robocorp_ls_core.lsp import MarkupContentTypedDict
from robocorp_ls_core.lsp import MarkupKind
from robocorp_ls_core.lsp import ParameterInformation
from robocorp_ls_core.lsp import SignatureHelp
from robocorp_ls_core.lsp import SignatureInformation
from robotframework_ls.impl.protocols import ICompletionContext, IKeywordFound
from robotframework_ls.impl.protocols import IKeywordDefinition
from robotframework_ls.impl.protocols import KeywordUsageInfo

from typing import List, Optional


def _library_signature_help(
    completion_context: ICompletionContext, library_node
) -> Optional[SignatureHelp]:
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.robot_specbuilder import docs_and_format
    from robot.api import Token
    from robotframework_ls.impl.keyword_argument_analysis import KeywordArgumentAnalysis
    from robotframework_ls.impl.keyword_argument_analysis import (
        UsageInfoForKeywordArgumentAnalysis,
    )
    from robotframework_ls.impl.libspec_manager import LibspecManager

    name = library_node.name
    if not name:
        return None

    libspec_manager: LibspecManager = completion_context.workspace.libspec_manager
    args = ast_utils.get_library_arguments_serialized(library_node)

    library_doc = libspec_manager.get_library_doc_or_error(
        name,
        create=True,
        completion_context=completion_context,
        builtin=False,
        args=args,
    ).library_doc
    if library_doc is None:
        return None

    docs, docs_format = docs_and_format(library_doc)
    documentation: MarkupContentTypedDict = {
        "kind": MarkupKind.Markdown
        if docs_format == "markdown"
        else MarkupKind.PlainText,
        "value": docs,
    }

    arg_names_as_list: List[str] = []
    active_parameter = -1
    parameters: List[ParameterInformation] = []
    for keyword_doc in library_doc.inits:
        for arg in keyword_doc.args:
            arg_names_as_list.append(arg.original_arg)
            parameters.append(
                ParameterInformation(arg.original_arg, None),
            )

        name_token = library_node.get_token(Token.NAME)
        if name_token is not None:
            keyword_analysis = KeywordArgumentAnalysis(keyword_doc.args)
            active_parameter = keyword_analysis.compute_active_parameter(
                UsageInfoForKeywordArgumentAnalysis(library_node, name_token),
                lineno=completion_context.sel.line,
                col=completion_context.sel.col,
            )

        break  # Just add the first one...

    if not arg_names_as_list:
        label = library_doc.name
    else:
        label = "%s(%s)" % (library_doc.name, ", ".join(arg_names_as_list))

    signatures: List[SignatureInformation] = [
        SignatureInformation(label, documentation, parameters)
    ]
    ret = SignatureHelp(
        signatures, active_signature=0, active_parameter=active_parameter
    )
    ret.name = library_doc.name
    ret.node = library_node
    return ret


def signature_help_internal(
    completion_context: ICompletionContext,
) -> Optional[SignatureHelp]:
    token_info = completion_context.get_current_token()
    if token_info is None:
        return None

    if isinstance_name(token_info.node, "LibraryImport"):
        return _library_signature_help(completion_context, token_info.node)

    keyword_definition_and_usage_info = (
        completion_context.get_current_keyword_definition_and_usage_info()
    )
    if keyword_definition_and_usage_info is None:
        return None

    from robotframework_ls.impl.keyword_argument_analysis import KeywordArgumentAnalysis
    from robotframework_ls.impl.keyword_argument_analysis import (
        UsageInfoForKeywordArgumentAnalysis,
    )
    from robot.api import Token

    keyword_definition: IKeywordDefinition = keyword_definition_and_usage_info[0]
    usage_info: KeywordUsageInfo = keyword_definition_and_usage_info[1]
    keyword_found: IKeywordFound = keyword_definition.keyword_found

    keyword_args = keyword_found.keyword_args
    keyword_analysis = KeywordArgumentAnalysis(keyword_args)

    keyword_token = usage_info.node.get_token(Token.KEYWORD)

    if keyword_token is None:
        active_parameter = -1
    else:
        active_parameter = keyword_analysis.compute_active_parameter(
            UsageInfoForKeywordArgumentAnalysis(usage_info.node, keyword_token),
            lineno=completion_context.sel.line,
            col=completion_context.sel.col,
        )

    documentation = keyword_found.compute_docs_without_signature()

    arg_names_as_list: List[str] = []
    parameters: List[ParameterInformation] = []

    for arg in keyword_args:
        arg_names_as_list.append(arg.original_arg)
        parameters.append(
            ParameterInformation(arg.original_arg, None),
        )

    if not arg_names_as_list:
        label = keyword_found.keyword_name
    else:
        label = "%s(%s)" % (keyword_found.keyword_name, ", ".join(arg_names_as_list))

    signatures: List[SignatureInformation] = [
        SignatureInformation(label, documentation, parameters)
    ]
    ret = SignatureHelp(
        signatures, active_signature=0, active_parameter=active_parameter
    )
    ret.name = keyword_found.keyword_name
    ret.node = usage_info.node
    return ret


def signature_help(completion_context: ICompletionContext) -> Optional[dict]:
    ret = signature_help_internal(completion_context)
    if ret is None:
        return None

    return ret.to_dict()



from typing import List

from robotframework_ls.impl.protocols import ICompletionContext
from robocorp_ls_core.lsp import DocumentSymbolTypedDict, SymbolKind


def collect_children(ast) -> List[DocumentSymbolTypedDict]:
    from robotframework_ls.impl import ast_utils
    from robot.api import Token
    from robotframework_ls.impl.ast_utils import create_range_from_token

    ret: List[DocumentSymbolTypedDict] = []

    for node_info in ast_utils.iter_nodes(
        ast, accept_class=("Keyword", "TestCase", "Variable")
    ):
        node = node_info.node
        classname = node.__class__.__name__
        if classname == "Keyword":
            token = node.header.get_token(Token.KEYWORD_NAME)
            symbol_range = create_range_from_token(token)
            doc_symbol: DocumentSymbolTypedDict = {
                "name": str(token),
                "kind": SymbolKind.Function,
                "range": symbol_range,
                "selectionRange": symbol_range,
            }
            ret.append(doc_symbol)

        elif classname == "TestCase":
            token = node.header.get_token(Token.TESTCASE_NAME)
            symbol_range = create_range_from_token(token)
            doc_symbol = {
                "name": str(token),
                "kind": SymbolKind.Class,
                "range": symbol_range,
                "selectionRange": symbol_range,
            }
            ret.append(doc_symbol)

        elif classname == "Variable":
            token = node.get_token(Token.VARIABLE)
            symbol_range = create_range_from_token(token)
            doc_symbol = {
                "name": str(token),
                "kind": SymbolKind.Variable,
                "range": symbol_range,
                "selectionRange": symbol_range,
            }
            ret.append(doc_symbol)

    return ret


def create_section_doc_symbol(
    ret: List[DocumentSymbolTypedDict], ast, header_token_type, symbol_kind
):
    from robotframework_ls.impl.ast_utils import create_range_from_token

    token = ast.header.get_token(header_token_type)
    if token is not None:
        symbol_range = create_range_from_token(token)
        doc_symbol: DocumentSymbolTypedDict = {
            "name": str(token).replace("*", "").strip(),
            "kind": symbol_kind,
            "range": symbol_range,
            "selectionRange": symbol_range,
            "children": collect_children(ast),
        }
        ret.append(doc_symbol)


def document_symbol(
    completion_context: ICompletionContext,
) -> List[DocumentSymbolTypedDict]:
    from robotframework_ls.impl import ast_utils
    from robot.api import Token

    ret: List[DocumentSymbolTypedDict] = []
    ast = completion_context.get_ast()

    for node_info in ast_utils.iter_nodes(ast, "SettingSection"):
        create_section_doc_symbol(
            ret, node_info.node, Token.SETTING_HEADER, SymbolKind.Namespace
        )

    for node_info in ast_utils.iter_nodes(ast, "VariableSection"):
        create_section_doc_symbol(
            ret, node_info.node, Token.VARIABLE_HEADER, SymbolKind.Namespace
        )

    for node_info in ast_utils.iter_nodes(ast, "TestCaseSection"):
        create_section_doc_symbol(
            ret, node_info.node, Token.TESTCASE_HEADER, SymbolKind.Namespace
        )

    for node_info in ast_utils.iter_nodes(ast, "KeywordSection"):
        create_section_doc_symbol(
            ret, node_info.node, Token.KEYWORD_HEADER, SymbolKind.Namespace
        )

    return ret



from _pydevd_bundle.pydevd_extension_api import DebuggerEventHandler
import os
import sys


class VerifyEvent(object):
    def on_debugger_modules_loaded(self, **kwargs):
        print ("INITIALIZE EVENT RECEIVED")
        # check that some core modules are loaded before this callback is invoked
        modules_loaded = all(mod in sys.modules for mod in ('pydevd_file_utils', '_pydevd_bundle.pydevd_constants'))
        if modules_loaded:
            print ("TEST SUCEEDED")  # incorrect spelling on purpose
        else:
            print ("TEST FAILED")


if os.environ.get("VERIFY_EVENT_TEST"):
    DebuggerEventHandler.register(VerifyEvent)



"""
This is done to solve the use-case where we can (potentially) have a different
interpreter for different directories.

i.e.:

Given some structure as:

/env1
    /robot.yaml <- specifies libraries needed and additional pythonpath entries.
    /robot1.robot
/env2
    /robot.yaml
    /robot2.robot
    
We want to use the info on robot.yaml in env1 to say that a given set of
libraries is needed, whereas the robot.yaml in env2 has another set of 
libraries (and each will have a different virtual environment managing it).

Note that this implementation logic is not really builtin, rather, an external
contributor needs to add the extension point to do that.

The usage of the extension point should do something as:

    interpreter_info = ep.get_interpreter_info_for_doc_uri(doc_uri) 
    if interpreter_info is not None:
        interpreter_info.get_interpreter_id()
        interpreter_info.get_python_exe()
        interpreter_info.get_environ()
        
Later, it can decide to get rid of some server which handles a given python
executable if there's no interpreter id using a given python executable anymore.

Note: to add an extension, one needs to add it through a plugin using the 
"robot.addPluginsDir" command.
"""

from typing import Optional, List, Dict
import sys

# Hack so that we don't break the runtime on versions prior to Python 3.8.
if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass


else:
    from typing import Protocol


class IInterpreterInfo(Protocol):
    def get_interpreter_id(self) -> str:
        """
        This is an identification and should relate to the place which allows
        to identify this info (it should usually be given by some location --
        for instance, it could be identified by the robot.yaml location).

        Note that different interpreter ids can point to the same python
        executable (if they'd have the same robot.yaml contents).
        """

    def get_python_exe(self) -> str:
        """
        The python executable that should be used.
        """

    def get_environ(self) -> Optional[Dict[str, str]]:
        """
        The environment to be used.
        """

    def get_additional_pythonpath_entries(self) -> List[str]:
        """
        Any additional PYTHONPATH entries to be considered.
        """


class EPResolveInterpreter(Protocol):
    def get_interpreter_info_for_doc_uri(self, doc_uri) -> Optional[IInterpreterInfo]:
        """
        Provides a customized interpreter for a given document uri.
        """


class DefaultInterpreterInfo(object):
    """
    A Default implementation for the interpreter info where everything is
    pre-computed.
    """

    def __init__(
        self,
        interpreter_id: str,
        python_exe: str,
        environ: Optional[Dict[str, str]],
        additional_pythonpath_entries: List[str],
    ) -> None:
        self.interpreter_id = interpreter_id
        self.python_exe = python_exe
        self.environ = environ
        self.additional_pythonpath_entries = additional_pythonpath_entries

    def get_interpreter_id(self) -> str:
        return self.interpreter_id

    def get_python_exe(self) -> str:
        return self.python_exe

    def get_environ(self) -> Optional[Dict[str, str]]:
        return self.environ

    def get_additional_pythonpath_entries(self) -> List[str]:
        return self.additional_pythonpath_entries

    def __str__(self):
        return f"DefaultInterpreterInfo({self.interpreter_id}, {self.python_exe})"

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IInterpreterInfo = check_implements(self)



import os.path
from robocorp_ls_core.robotframework_log import get_logger
from typing import Optional, List
from robotframework_ls.impl.protocols import ICompletionContext
from robocorp_ls_core.lsp import CompletionItemTypedDict
from robocorp_ls_core.basic import normalize_filename
from robotframework_ls.impl.robot_constants import (
    ROBOT_AND_TXT_FILE_EXTENSIONS,
    LIBRARY_FILE_EXTENSIONS,
)

log = get_logger(__name__)


def _create_completion_item(
    library_name, selection, token, start_col_offset=None
) -> CompletionItemTypedDict:
    from robocorp_ls_core.lsp import (
        CompletionItem,
        InsertTextFormat,
        Position,
        Range,
        TextEdit,
    )
    from robocorp_ls_core.lsp import CompletionItemKind

    text_edit = TextEdit(
        Range(
            start=Position(
                selection.line,
                start_col_offset if start_col_offset is not None else token.col_offset,
            ),
            end=Position(selection.line, token.end_col_offset),
        ),
        library_name,
    )

    # text_edit = None
    return CompletionItem(
        library_name,
        kind=CompletionItemKind.Module,
        text_edit=text_edit,
        insertText=text_edit.newText,
        documentation="",
        insertTextFormat=InsertTextFormat.Snippet,
    ).to_dict()


def _add_completions_from_dir(
    completion_context,
    directory,
    matcher,
    ret: List[CompletionItemTypedDict],
    sel,
    token,
    qualifier,
    extensions,
    skip_current,
):
    from robocorp_ls_core import uris

    def normfile(path):
        return normalize_filename(path)

    curr_file = normfile(uris.to_fs_path(completion_context.doc.uri))

    try:
        # This is ok if the directory doesn't exist.
        contents = sorted(os.listdir(directory))
    except:
        return

    for filename in contents:
        use_path = None
        if filename.endswith(extensions):
            # If that'd be a match for the current .robot file, don't show it.
            if skip_current and curr_file == normfile(
                os.path.join(directory, filename)
            ):
                continue

            use_path = filename

        elif filename not in ("__pycache__", ".git") and os.path.isdir(
            os.path.join(directory, filename)
        ):
            use_path = filename + "/"
        else:
            continue

        if matcher.accepts(use_path):

            ret.append(
                _create_completion_item(
                    use_path, sel, token, start_col_offset=sel.col - len(qualifier)
                )
            )


def _get_completions(
    completion_context, token, match_libs, extensions, skip_current
) -> List[CompletionItemTypedDict]:
    from robotframework_ls.impl.string_matcher import RobotStringMatcher
    from robocorp_ls_core import uris
    from robotframework_ls.impl.robot_constants import BUILTIN_LIB, RESERVED_LIB
    from robotframework_ls.impl import ast_utils

    ret: List[CompletionItemTypedDict] = []

    sel = completion_context.sel
    value_to_cursor = token.value
    if token.end_col_offset > sel.col:
        value_to_cursor = value_to_cursor[: -(token.end_col_offset - sel.col)]
    if "{" in value_to_cursor:
        value_to_cursor = completion_context.token_value_resolving_variables(
            ast_utils.create_token(value_to_cursor)
        )

    value_to_cursor_split = os.path.split(value_to_cursor)

    if os.path.isabs(value_to_cursor):
        _add_completions_from_dir(
            completion_context,
            value_to_cursor_split[0],
            RobotStringMatcher(value_to_cursor_split[1]),
            ret,
            sel,
            token,
            value_to_cursor_split[1],
            extensions,
            skip_current=skip_current,
        )

    else:
        if match_libs:
            matcher = RobotStringMatcher(value_to_cursor)
            libspec_manager = completion_context.workspace.libspec_manager
            library_names = set(libspec_manager.get_library_names())
            library_names.discard(BUILTIN_LIB)
            library_names.discard(RESERVED_LIB)

            for library_name in library_names:
                if matcher.accepts(library_name):
                    ret.append(_create_completion_item(library_name, sel, token))

        # After checking the existing library names in memory (because we
        # loaded them at least once), check libraries in the filesystem.
        uri = completion_context.doc.uri
        path = uris.to_fs_path(uri)
        dirname = os.path.dirname(path)

        matcher = RobotStringMatcher(value_to_cursor_split[1])
        directory = os.path.join(dirname, value_to_cursor_split[0])
        _add_completions_from_dir(
            completion_context,
            directory,
            matcher,
            ret,
            sel,
            token,
            value_to_cursor_split[1],
            extensions,
            skip_current=skip_current,
        )
    return ret


def _get_resource_completions(
    completion_context, token
) -> List[CompletionItemTypedDict]:
    return _get_completions(
        completion_context,
        token,
        False,
        ROBOT_AND_TXT_FILE_EXTENSIONS,
        skip_current=True,
    )


def _get_library_completions(
    completion_context, token
) -> List[CompletionItemTypedDict]:
    return _get_completions(
        completion_context, token, True, LIBRARY_FILE_EXTENSIONS, skip_current=False
    )


class _Requisites(object):
    def __init__(self, token, is_library: bool):
        self.token = token
        self.is_library = is_library


def get_requisites(completion_context: ICompletionContext) -> Optional[_Requisites]:
    from robotframework_ls.impl import ast_utils

    token_info = completion_context.get_current_token()
    if token_info is not None:
        token = ast_utils.get_library_import_name_token(
            token_info.node, token_info.token
        )
        if token is not None:
            return _Requisites(token, True)
        else:
            token = ast_utils.get_resource_import_name_token(
                token_info.node, token_info.token
            )
            if token is not None:
                return _Requisites(token, False)
    return None


def complete(completion_context: ICompletionContext) -> List[CompletionItemTypedDict]:
    """
    Provides the completions for 'Library' and 'Resource' imports.
    """
    try:

        requisites = get_requisites(completion_context)
        if requisites is None:
            return []

        if requisites.is_library:
            return _get_library_completions(completion_context, requisites.token)
        else:
            return _get_resource_completions(completion_context, requisites.token)

    except:
        log.exception()

    return []



import re
from typing import Optional

from robot.api.parsing import ModelTransformer, Token

from robotidy.decorators import check_start_end_line
from robotidy.exceptions import InvalidParameterValueError


class RenameTestCases(ModelTransformer):
    r"""
    Enforce test case naming.

    Capitalize first letter of test case name, remove trailing dot and strip leading/trailing whitespace.

    It is also possible to configure `replace_pattern` parameter to find and replace regex pattern. Use `replace_to`
    to set replacement value. This configuration:

        robotidy --transform RenameTestCases -c RenameTestCases:replace_pattern=[A-Z]{3,}-\d{2,}:replace_to=foo

    will transform following code:

        *** Test Cases ***
        test ABC-123
            No Operation

    To:

        *** Test Cases ***
        Test foo
            No Operation

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/RenameTestCases.html for more examples.
    """
    ENABLED = False

    def __init__(self, replace_pattern: Optional[str] = None, replace_to: Optional[str] = None):
        try:
            self.replace_pattern = re.compile(replace_pattern) if replace_pattern is not None else None
        except re.error as err:
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "replace_pattern",
                replace_pattern,
                f"It should be a valid regex expression. Regex error: '{err.msg}'",
            )
        self.replace_to = "" if replace_to is None else replace_to

    @check_start_end_line
    def visit_TestCaseName(self, node):  # noqa
        token = node.get_token(Token.TESTCASE_NAME)
        if token.value:
            token.value = token.value[0].upper() + token.value[1:]
            if self.replace_pattern is not None:
                token.value = self.replace_pattern.sub(repl=self.replace_to, string=token.value)
            if token.value.endswith("."):
                token.value = token.value[:-1]
            token.value = token.value.strip()
        return node



# Original work Copyright 2017 Palantir Technologies, Inc. (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.basic import implements
from robocorp_ls_core.protocols import IConfig, Sentinel
from typing import Any, FrozenSet, Optional
import os


log = get_logger(__name__)


def flatten_keys(d: dict, parent_key="", all_options=frozenset(), result_dict=None):
    if result_dict is None:
        result_dict = {}

    for k, v in d.items():
        new_key = parent_key + "." + k if parent_key else k

        if new_key not in all_options and isinstance(v, dict):
            flatten_keys(v, new_key, all_options, result_dict)
            continue

        result_dict[new_key] = v
    return result_dict


class Config(object):
    ALL_OPTIONS: FrozenSet[str] = frozenset()

    def __init__(self, all_options: FrozenSet[str] = frozenset()):
        if all_options:
            self.ALL_OPTIONS = all_options

        self._settings: dict = {}
        self._override_settings: dict = {}

        self._original_settings: dict = {}
        self._original_override_settings: dict = {}

        self._full_settings: dict = {}
        self._workspace_dir: Optional[str] = None

    @implements(IConfig.get_setting)
    def get_setting(self, key, expected_type, default=Sentinel.SENTINEL) -> Any:
        try:
            s = self._full_settings[key]
            if not isinstance(s, expected_type):
                try:
                    # Check if we can cast it...
                    s = expected_type(s)
                except:
                    raise KeyError(
                        "Expected %s to be a setting of type: %s. Found: %s"
                        % (key, expected_type, type(s))
                    )
        except KeyError:
            if default is not Sentinel.SENTINEL:
                return default
            raise
        return s

    def _update_full_settings(self):
        full_settings = self._settings.copy()
        full_settings.update(self._override_settings)
        self._full_settings = full_settings
        log.debug("Updated settings to %s", full_settings)

    def _get_var_value(self, name):
        ret = name
        if name in ("${workspace}", "${workspaceRoot}", "${workspaceFolder}"):
            if self._workspace_dir is not None:
                ret = self._workspace_dir
            else:
                log.info("Unable to make workspace replacement for variable: %s", name)

        elif (name.startswith("${env.") or name.startswith("${env:")) and name.endswith(
            "}"
        ):
            name = name[6:-1]
            ret = os.environ.get(name)  # Note: should be case-insensitive on windows.
        else:
            log.info("Unable to resolve variable: %s", name)

        return ret

    def _var_replace(self, option, value):
        import re

        compiled = re.compile(r"\${([^{}]*)}")
        lasti = 0
        new_value = []
        for o in compiled.finditer(value):
            new_value.append(value[lasti : o.start()])
            new_value.append(self._get_var_value(o.group(0)))
            lasti = o.end()

        if lasti == 0:
            # Nothing changed
            return value

        new_value.append(value[lasti:])
        ret = "".join(new_value)
        if ret.startswith("~"):
            ret = os.path.expanduser(ret)

        log.debug("Changed setting: %s from %s to %s", option, value, ret)

        return ret

    def _replace_variables_in_settings(self, settings: dict) -> dict:
        """
        :param settings:
            The settings where the variables should be replaced.
            Note that this instance is unchanged.

        :return dict:
            Returns a new dict with the variables replaced.
        """
        settings = settings.copy()
        for option in self.ALL_OPTIONS:
            value = settings.get(option)
            if isinstance(value, str):
                settings[option] = self._var_replace(option, value)

            elif isinstance(value, list):
                new_value = []
                for val in value:
                    if isinstance(val, str):
                        new_value.append(self._var_replace(option, val))
                    else:
                        new_value.append(val)
                settings[option] = new_value

            elif isinstance(value, dict):
                new_dct = {}
                for key, val in value.items():
                    if isinstance(val, str):
                        new_dct[key] = self._var_replace(option, val)
                    else:
                        new_dct[key] = val
                settings[option] = new_dct
        return settings

    @implements(IConfig.update)
    def update(self, settings: dict):
        settings = flatten_keys(settings, all_options=self.ALL_OPTIONS)
        self._original_settings = settings
        self._settings = self._replace_variables_in_settings(settings)
        self._update_full_settings()

    @implements(IConfig.set_override_settings)
    def set_override_settings(self, override_settings):
        settings = flatten_keys(override_settings, all_options=self.ALL_OPTIONS)
        self._original_override_settings = settings
        self._override_settings = self._replace_variables_in_settings(settings)
        self._update_full_settings()

    @implements(IConfig.update_override_settings)
    def update_override_settings(self, override_settings):
        settings = flatten_keys(override_settings, all_options=self.ALL_OPTIONS)
        original = self._original_override_settings.copy()
        original.update(settings)

        self._original_override_settings = original
        self._override_settings = self._replace_variables_in_settings(original)
        self._update_full_settings()

    @implements(IConfig.get_full_settings)
    def get_full_settings(self):
        return self._full_settings

    @implements(IConfig.set_workspace_dir)
    def set_workspace_dir(self, workspace: str):
        self._workspace_dir = workspace
        self._settings = self._replace_variables_in_settings(self._original_settings)
        self._override_settings = self._replace_variables_in_settings(
            self._original_override_settings
        )
        self._update_full_settings()



import time
import sys
import itertools

from itertools import groupby
count = itertools.count(0)


def next_val():
    return next(count) % 25


start_time = time.time()
letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

# create an array of random strings of 40 characters each
l = sorted([''.join([letters[next_val()] for _ in range(40)]) for _ in range(10000)])
# group by the first two characters
g = {k: list(v) for k, v in groupby(l, lambda x: x[:2])}

if False:
    pass  # Breakpoint here

print('TotalTime>>%s<<' % (time.time() - start_time,))
print('TEST SUCEEDED')



import unittest

class SampleTest(unittest.TestCase):
    
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_xxxxxx1(self):
        self.fail('Fail test 2')
    def test_xxxxxx2(self):
        pass
    def test_xxxxxx3(self):
        pass
    def test_xxxxxx4(self):
        pass
    def test_non_unique_name(self):
        print('non unique name ran')


class AnotherSampleTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_1(self):
        pass
    def test_2(self):
        """ im a doc string"""
        pass
    def todo_not_tested(self):
        '''
        Not there by default!
        '''


if __name__ == '__main__':
#    suite = unittest.makeSuite(SampleTest, 'test')
#    runner = unittest.TextTestRunner( verbosity=3 )
#    runner.run(suite)
    unittest.main()



"""Utilities needed to emulate Python's interactive interpreter.

"""

# Inspired by similar code by Jeff Epler and Fredrik Lundh.

import sys
import traceback

# START --------------------------- from codeop import CommandCompiler, compile_command
# START --------------------------- from codeop import CommandCompiler, compile_command
# START --------------------------- from codeop import CommandCompiler, compile_command
# START --------------------------- from codeop import CommandCompiler, compile_command
# START --------------------------- from codeop import CommandCompiler, compile_command
r"""Utilities to compile possibly incomplete Python source code.

This module provides two interfaces, broadly similar to the builtin
function compile(), which take program text, a filename and a 'mode'
and:

- Return code object if the command is complete and valid
- Return None if the command is incomplete
- Raise SyntaxError, ValueError or OverflowError if the command is a
  syntax error (OverflowError and ValueError can be produced by
  malformed literals).

Approach:

First, check if the source consists entirely of blank lines and
comments; if so, replace it with 'pass', because the built-in
parser doesn't always do the right thing for these.

Compile three times: as is, with \n, and with \n\n appended.  If it
compiles as is, it's complete.  If it compiles with one \n appended,
we expect more.  If it doesn't compile either way, we compare the
error we get when compiling with \n or \n\n appended.  If the errors
are the same, the code is broken.  But if the errors are different, we
expect more.  Not intuitive; not even guaranteed to hold in future
releases; but this matches the compiler's behavior from Python 1.4
through 2.2, at least.

Caveat:

It is possible (but not likely) that the parser stops parsing with a
successful outcome before reaching the end of the source; in this
case, trailing symbols may be ignored instead of causing an error.
For example, a backslash followed by two newlines may be followed by
arbitrary garbage.  This will be fixed once the API for the parser is
better.

The two interfaces are:

compile_command(source, filename, symbol):

    Compiles a single command in the manner described above.

CommandCompiler():

    Instances of this class have __call__ methods identical in
    signature to compile_command; the difference is that if the
    instance compiles program text containing a __future__ statement,
    the instance 'remembers' and compiles all subsequent program texts
    with the statement in force.

The module also provides another class:

Compile():

    Instances of this class act like the built-in function compile,
    but with 'memory' in the sense described above.
"""

import __future__

_features = [getattr(__future__, fname)
             for fname in __future__.all_feature_names]

__all__ = ["compile_command", "Compile", "CommandCompiler"]

PyCF_DONT_IMPLY_DEDENT = 0x200  # Matches pythonrun.h


def _maybe_compile(compiler, source, filename, symbol):
    # Check for source consisting of only blank lines and comments
    for line in source.split("\n"):
        line = line.strip()
        if line and line[0] != '#':
            break  # Leave it alone
    else:
        if symbol != "eval":
            source = "pass"  # Replace it with a 'pass' statement

    err = err1 = err2 = None
    code = code1 = code2 = None

    try:
        code = compiler(source, filename, symbol)
    except SyntaxError as err:
        pass

    try:
        code1 = compiler(source + "\n", filename, symbol)
    except SyntaxError as err1:
        pass

    try:
        code2 = compiler(source + "\n\n", filename, symbol)
    except SyntaxError as err2:
        pass

    if code:
        return code
    if not code1 and repr(err1) == repr(err2):
        raise SyntaxError(err1)


def _compile(source, filename, symbol):
    return compile(source, filename, symbol, PyCF_DONT_IMPLY_DEDENT)


def compile_command(source, filename="<input>", symbol="single"):
    r"""Compile a command and determine whether it is incomplete.

    Arguments:

    source -- the source string; may contain \n characters
    filename -- optional filename from which source was read; default
                "<input>"
    symbol -- optional grammar start symbol; "single" (default) or "eval"

    Return value / exceptions raised:

    - Return a code object if the command is complete and valid
    - Return None if the command is incomplete
    - Raise SyntaxError, ValueError or OverflowError if the command is a
      syntax error (OverflowError and ValueError can be produced by
      malformed literals).
    """
    return _maybe_compile(_compile, source, filename, symbol)


class Compile:
    """Instances of this class behave much like the built-in compile
    function, but if one is used to compile text containing a future
    statement, it "remembers" and compiles all subsequent program texts
    with the statement in force."""

    def __init__(self):
        self.flags = PyCF_DONT_IMPLY_DEDENT

    def __call__(self, source, filename, symbol):
        codeob = compile(source, filename, symbol, self.flags, 1)
        for feature in _features:
            if codeob.co_flags & feature.compiler_flag:
                self.flags |= feature.compiler_flag
        return codeob


class CommandCompiler:
    """Instances of this class have __call__ methods identical in
    signature to compile_command; the difference is that if the
    instance compiles program text containing a __future__ statement,
    the instance 'remembers' and compiles all subsequent program texts
    with the statement in force."""

    def __init__(self,):
        self.compiler = Compile()

    def __call__(self, source, filename="<input>", symbol="single"):
        r"""Compile a command and determine whether it is incomplete.

        Arguments:

        source -- the source string; may contain \n characters
        filename -- optional filename from which source was read;
                    default "<input>"
        symbol -- optional grammar start symbol; "single" (default) or
                  "eval"

        Return value / exceptions raised:

        - Return a code object if the command is complete and valid
        - Return None if the command is incomplete
        - Raise SyntaxError, ValueError or OverflowError if the command is a
          syntax error (OverflowError and ValueError can be produced by
          malformed literals).
        """
        return _maybe_compile(self.compiler, source, filename, symbol)

# END --------------------------- from codeop import CommandCompiler, compile_command
# END --------------------------- from codeop import CommandCompiler, compile_command
# END --------------------------- from codeop import CommandCompiler, compile_command
# END --------------------------- from codeop import CommandCompiler, compile_command
# END --------------------------- from codeop import CommandCompiler, compile_command


__all__ = ["InteractiveInterpreter", "InteractiveConsole", "interact",
           "compile_command"]


def softspace(file, newvalue):
    oldvalue = 0
    try:
        oldvalue = file.softspace
    except AttributeError:
        pass
    try:
        file.softspace = newvalue
    except (AttributeError, TypeError):
        # "attribute-less object" or "read-only attributes"
        pass
    return oldvalue


class InteractiveInterpreter:
    """Base class for InteractiveConsole.

    This class deals with parsing and interpreter state (the user's
    namespace); it doesn't deal with input buffering or prompting or
    input file naming (the filename is always passed in explicitly).

    """

    def __init__(self, locals=None):
        """Constructor.

        The optional 'locals' argument specifies the dictionary in
        which code will be executed; it defaults to a newly created
        dictionary with key "__name__" set to "__console__" and key
        "__doc__" set to None.

        """
        if locals is None:
            locals = {"__name__": "__console__", "__doc__": None}
        self.locals = locals
        self.compile = CommandCompiler()

    def runsource(self, source, filename="<input>", symbol="single"):
        """Compile and run some source in the interpreter.

        Arguments are as for compile_command().

        One several things can happen:

        1) The input is incorrect; compile_command() raised an
        exception (SyntaxError or OverflowError).  A syntax traceback
        will be printed by calling the showsyntaxerror() method.

        2) The input is incomplete, and more input is required;
        compile_command() returned None.  Nothing happens.

        3) The input is complete; compile_command() returned a code
        object.  The code is executed by calling self.runcode() (which
        also handles run-time exceptions, except for SystemExit).

        The return value is True in case 2, False in the other cases (unless
        an exception is raised).  The return value can be used to
        decide whether to use sys.ps1 or sys.ps2 to prompt the next
        line.

        """
        try:
            code = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            # Case 1
            self.showsyntaxerror(filename)
            return False

        if code is None:
            # Case 2
            return True

        # Case 3
        self.runcode(code)
        return False

    def runcode(self, code):
        """Execute a code object.

        When an exception occurs, self.showtraceback() is called to
        display a traceback.  All exceptions are caught except
        SystemExit, which is reraised.

        A note about KeyboardInterrupt: this exception may occur
        elsewhere in this code, and may not always be caught.  The
        caller should be prepared to deal with it.

        """
        try:
            exec(code, self.locals)
        except SystemExit:
            raise
        except:
            self.showtraceback()
        else:
            if softspace(sys.stdout, 0):
                sys.stdout.write('\n')

    def showsyntaxerror(self, filename=None):
        """Display the syntax error that just occurred.

        This doesn't display a stack trace because there isn't one.

        If a filename is given, it is stuffed in the exception instead
        of what was there before (because Python's parser always uses
        "<string>" when reading from a string).

        The output is written by self.write(), below.

        """
        type, value, sys.last_traceback = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        if filename and type is SyntaxError:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (dummy_filename, lineno, offset, line) = value
            except:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename
                value = SyntaxError(msg, (filename, lineno, offset, line))
                sys.last_value = value
        list = traceback.format_exception_only(type, value)
        map(self.write, list)

    def showtraceback(self, *args, **kwargs):
        """Display the exception that just occurred.

        We remove the first stack item because it is our own code.

        The output is written by self.write(), below.

        """
        try:
            type, value, tb = sys.exc_info()
            sys.last_type = type
            sys.last_value = value
            sys.last_traceback = tb
            tblist = traceback.extract_tb(tb)
            del tblist[:1]
            list = traceback.format_list(tblist)
            if list:
                list.insert(0, "Traceback (most recent call last):\n")
            list[len(list):] = traceback.format_exception_only(type, value)
        finally:
            tblist = tb = None
        map(self.write, list)

    def write(self, data):
        """Write a string.

        The base implementation writes to sys.stderr; a subclass may
        replace this with a different implementation.

        """
        sys.stderr.write(data)


class InteractiveConsole(InteractiveInterpreter):
    """Closely emulate the behavior of the interactive Python interpreter.

    This class builds on InteractiveInterpreter and adds prompting
    using the familiar sys.ps1 and sys.ps2, and input buffering.

    """

    def __init__(self, locals=None, filename="<console>"):
        """Constructor.

        The optional locals argument will be passed to the
        InteractiveInterpreter base class.

        The optional filename argument should specify the (file)name
        of the input stream; it will show up in tracebacks.

        """
        InteractiveInterpreter.__init__(self, locals)
        self.filename = filename
        self.resetbuffer()

    def resetbuffer(self):
        """Reset the input buffer."""
        self.buffer = []

    def interact(self, banner=None):
        """Closely emulate the interactive Python console.

        The optional banner argument specify the banner to print
        before the first interaction; by default it prints a banner
        similar to the one printed by the real Python interpreter,
        followed by the current class name in parentheses (so as not
        to confuse this with the real interpreter -- since it's so
        close!).

        """
        try:
            sys.ps1  # @UndefinedVariable
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2  # @UndefinedVariable
        except AttributeError:
            sys.ps2 = "... "
        cprt = 'Type "help", "copyright", "credits" or "license" for more information.'
        if banner is None:
            self.write("Python %s on %s\n%s\n(%s)\n" %
                       (sys.version, sys.platform, cprt,
                        self.__class__.__name__))
        else:
            self.write("%s\n" % str(banner))
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2  # @UndefinedVariable
                else:
                    prompt = sys.ps1  # @UndefinedVariable
                try:
                    line = self.raw_input(prompt)
                    # Can be None if sys.stdin was redefined
                    encoding = getattr(sys.stdin, "encoding", None)
                    if encoding and not isinstance(line, str):
                        line = line.decode(encoding)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    more = self.push(line)
            except KeyboardInterrupt:
                self.write("\nKeyboardInterrupt\n")
                self.resetbuffer()
                more = 0

    def push(self, line):
        """Push a line to the interpreter.

        The line should not have a trailing newline; it may have
        internal newlines.  The line is appended to a buffer and the
        interpreter's runsource() method is called with the
        concatenated contents of the buffer as source.  If this
        indicates that the command was executed or invalid, the buffer
        is reset; otherwise, the command is incomplete, and the buffer
        is left as it was after the line was appended.  The return
        value is 1 if more input is required, 0 if the line was dealt
        with in some way (this is the same as runsource()).

        """
        self.buffer.append(line)
        source = "\n".join(self.buffer)
        more = self.runsource(source, self.filename)
        if not more:
            self.resetbuffer()
        return more

    def raw_input(self, prompt=""):
        """Write a prompt and read a line.

        The returned line does not include the trailing newline.
        When the user enters the EOF key sequence, EOFError is raised.

        The base implementation uses the built-in function
        raw_input(); a subclass may replace this with a different
        implementation.

        """
        return input(prompt)


def interact(banner=None, readfunc=None, local=None):
    """Closely emulate the interactive Python interpreter.

    This is a backwards compatible interface to the InteractiveConsole
    class.  When readfunc is not specified, it attempts to import the
    readline module to enable GNU readline if it is available.

    Arguments (all optional, all default to None):

    banner -- passed to InteractiveConsole.interact()
    readfunc -- if not None, replaces InteractiveConsole.raw_input()
    local -- passed to InteractiveInterpreter.__init__()

    """
    console = InteractiveConsole(local)
    if readfunc is not None:
        console.raw_input = readfunc
    else:
        try:
            import readline
        except ImportError:
            pass
    console.interact(banner)


if __name__ == '__main__':
    import pdb
    pdb.run("interact()\n")
